/*++

Copyright (c) Sydus.

Module:

    DataReader.h

Authors:

    Sharvil Nanavati (sharvil) 2006-12-24

--*/

#ifndef DATAREADER_H
#define DATAREADER_H

#include "base.h"
#include "ByteOrder.h"
#include "StringEncoding.h"

template <typename ByteOrder = BigEndian, typename StringEncoding = Length8Ascii>
class DataReader
{
	public:
		enum ESeekMode
		{
			KSeekAbsolute,
			KSeekRelative,
			KSeekReverse
		};

	public:
		explicit DataReader(const BString & source, BString::size_type offset = 0);
		~DataReader();

		int8 ReadInt8();
		int16 ReadInt16();
		int32 ReadInt32();

		uint8 ReadUint8();
		uint16 ReadUint16();
		uint32 ReadUint32();

		BString ReadBString(BString::size_type length);
		std::string ReadString(std::string::size_type length = std::string::npos);

		void Seek(BString::size_type offset, ESeekMode seekMode);

		bool IsEndOfData() const;
		BString::size_type Offset() const;
		BString::size_type BytesRemaining() const;

	private:
		const BString & iSource;
		BString::size_type iIndex;
};

template <typename T1, typename T2>
DataReader <T1, T2>::DataReader(const BString & source, BString::size_type offset)
	: iSource(source),
	  iIndex(offset)
{
	if(iIndex > iSource.GetLength())
		throw invalid_argument("[DataReader::DataReader]: overflow.");
}

template <typename T1, typename T2>
DataReader <T1, T2>::~DataReader()
{
}

template <typename T1, typename T2>
int8 DataReader <T1, T2>::ReadInt8()
{
	if(iIndex >= iSource.GetLength())
		throw invalid_argument("[DataReader::ReadInt8]: overflow.");

	return iSource[iIndex++];
}

template <typename T1, typename T2>
int16 DataReader <T1, T2>::ReadInt16()
{
	return T1::GetInt16(iSource, iIndex);
}

template <typename T1, typename T2>
int32 DataReader <T1, T2>::ReadInt32()
{
	return T1::GetInt32(iSource, iIndex);
}

template <typename T1, typename T2>
uint8 DataReader <T1, T2>::ReadUint8()
{
	if(iIndex >= iSource.GetLength())
		return 0;

	return iSource[iIndex++];
}

template <typename T1, typename T2>
uint16 DataReader <T1, T2>::ReadUint16()
{
	return T1::GetUint16(iSource, iIndex);
}

template <typename T1, typename T2>
uint32 DataReader <T1, T2>::ReadUint32()
{
	return T1::GetUint32(iSource, iIndex);
}

template <typename T1, typename T2>
BString DataReader <T1, T2>::ReadBString(BString::size_type length)
{
	if(length + iIndex > iSource.GetLength())
		throw invalid_argument("[DataReader::ReadBString]: data overrun.");

	BString ret = iSource.Mid(iIndex, length);
	iIndex += length;
	return ret;
}

template <typename T1, typename T2>
std::string DataReader <T1, T2>::ReadString(std::string::size_type length)
{
	return T2::DeserializeString(iSource, iIndex, length);
}

template <typename T1, typename T2>
void DataReader <T1, T2>::Seek(BString::size_type offset, ESeekMode seekMode)
{
	switch(seekMode)
	{
		case KSeekAbsolute:
			if(offset > iSource.GetLength())
				throw invalid_argument("[DataReader::Seek]: overflow.");

			iIndex = offset;
			break;

		case KSeekRelative:
			if(offset + iIndex > iSource.GetLength())
				throw invalid_argument("[DataReader::Seek]: overflow.");

			iIndex += offset;
			break;

		case KSeekReverse:
			if(iSource.GetLength() < offset)
				throw invalid_argument("[DataReader::Seek]: underflow.");

			iIndex = iSource.GetLength() - offset;
			break;

		default:
			throw invalid_argument(Format("[DataReader::Seek]: unknown seek mode %u.", seekMode));
	}
}

template <typename T1, typename T2>
bool DataReader <T1, T2>::IsEndOfData() const
{
	return (BytesRemaining() == 0);
}

template <typename T1, typename T2>
BString::size_type DataReader <T1, T2>::Offset() const
{
	return iIndex;
}

template <typename T1, typename T2>
BString::size_type DataReader <T1, T2>::BytesRemaining() const
{
	return (iSource.GetLength() - iIndex);
}

#endif
